/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.guard.configuration.internal;

import java.io.File;
import java.util.Properties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.openexchange.config.ConfigurationService;
import com.openexchange.config.DefaultInterests;
import com.openexchange.config.Interests;
import com.openexchange.config.cascade.ComposedConfigProperty;
import com.openexchange.config.cascade.ConfigProperty;
import com.openexchange.config.cascade.ConfigView;
import com.openexchange.config.cascade.ConfigViewFactory;
import com.openexchange.exception.OXException;
import com.openexchange.guard.configuration.GuardConfigurationService;
import com.openexchange.guard.configuration.GuardProperty;
import com.openexchange.guard.configuration.osgi.Services;
import com.openexchange.serverconfig.ServerConfigService;

/**
 * {@link GuardConfigurationServiceImpl}
 *
 * @author <a href="mailto:ioannis.chouklis@open-xchange.com">Ioannis Chouklis</a>
 */
public class GuardConfigurationServiceImpl implements GuardConfigurationService {

    private static final Logger logger = LoggerFactory.getLogger(GuardConfigurationServiceImpl.class);

    private final ConfigurationService configurationService;

    /**
     * All reloadable properties
     */
    private static final String[] PROPERTIES = new String[] { "all properties in file" };

    /**
     * Initialises a new {@link GuardConfigurationServiceImpl}.
     */
    public GuardConfigurationServiceImpl(ConfigurationService configurationService) {
        super();
        this.configurationService = configurationService;
        reloadConfiguration(configurationService);
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.config.Reloadable#reloadConfiguration(com.openexchange.config.ConfigurationService)
     */
    @Override
    public void reloadConfiguration(ConfigurationService configService) {
        // no-op
    }

    @Override
    public String getProperty(String property) {
        return getServerProperty(property, String.class, null);
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfiguration#getProperty(com.openexchange.guard.config.GuardProperty)
     */
    @Override
    public String getProperty(GuardProperty property) {
        return getServerProperty(property.getFQPropertyName(), String.class, property.getDefaultValue(String.class));
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.configuration.GuardConfigurationService#getPropertyFromFile(java.lang.String, java.lang.String)
     */
    @Override
    public String getPropertyFromFile(String fileName, String property) {
        Properties properties = configurationService.getFile(fileName);
        return properties.getProperty(property);
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfiguration#getIntProperty(com.openexchange.guard.config.GuardProperty)
     */
    @Override
    public int getIntProperty(GuardProperty property) {
        return getServerProperty(property.getFQPropertyName(), int.class, property.getDefaultValue(Integer.class));
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfiguration#getBooleanProperty(com.openexchange.guard.config.GuardProperty)
     */
    @Override
    public boolean getBooleanProperty(GuardProperty property) {
        return getServerProperty(property.getFQPropertyName(), boolean.class, property.getDefaultValue(Boolean.class));
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfigurationService#getProperty(com.openexchange.guard.config.GuardProperty, int, int)
     */
    @Override
    public String getProperty(GuardProperty property, int userId, int contextId) {
        return getProperty(property.getFQPropertyName(), userId, contextId, property.getDefaultValue(String.class));
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfigurationService#getIntProperty(com.openexchange.guard.config.GuardProperty, int, int)
     */
    @Override
    public int getIntProperty(GuardProperty property, int userId, int contextId) {
        return getIntProperty(property.getFQPropertyName(), userId, contextId, property.getDefaultValue(Integer.class));
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfigurationService#getBooleanProperty(com.openexchange.guard.config.GuardProperty, int, int)
     */
    @Override
    public boolean getBooleanProperty(GuardProperty property, int userId, int contextId) {
        return getBooleanProperty(property.getFQPropertyName(), userId, contextId, property.getDefaultValue(Boolean.class));
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfigurationService#getProperty(java.lang.String, int, int)
     */
    @Override
    public String getProperty(String property, int userId, int contextId, String defaultValue) {
        return getUserProperty(property, userId, contextId, String.class, defaultValue);
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfigurationService#getIntProperty(java.lang.String, int, int)
     */
    @Override
    public int getIntProperty(String property, int userId, int contextId, int defaultValue) {
        return getUserProperty(property, userId, contextId, int.class, defaultValue);
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.config.GuardConfigurationService#getBooleanProperty(java.lang.String, int, int)
     */
    @Override
    public boolean getBooleanProperty(String property, int userId, int contextId, boolean defaultValue) {
        return getUserProperty(property, userId, contextId, boolean.class, defaultValue);
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.configuration.GuardConfigurationService#getPropertyFile(java.lang.String)
     */
    @Override
    public File getPropertyFile(String propertyFileName) {
        return configurationService.getFileByName(propertyFileName);
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.configuration.GuardConfigurationService#getPropertyDirectory(java.lang.String)
     */
    @Override
    public File getPropertyDirectory(String propertyDirectory) {
        return configurationService.getDirectory(propertyDirectory);
    }

    @Override
    public <T> T getServerConfiguration(String host, String configKey) throws OXException {
        return getServerConfiguration(host, 0, 0, configKey);
    }

    @SuppressWarnings("unchecked")
    @Override
    public <T> T getServerConfiguration(String host, int userid, int cid, String configKey) throws OXException {
        ServerConfigService serverConfig = Services.getService(ServerConfigService.class);
        com.openexchange.serverconfig.ServerConfig config = serverConfig.getServerConfig(host, userid, cid);
        return (T) config.asMap().get(configKey);
    }

    @Override
	public String getContextSetStringProperty(GuardProperty property, int userId, int contextId, String defaultValue) throws OXException {
    	return getContextSetProperty(property.getFQPropertyName(), userId, contextId, String.class, defaultValue);
    }

    /**
     * Gets the value T of specified server property and coerce it to the specified type T
     *
     * @param property The name of the property
     * @param coerceTo The type T to coerce the value of the property to
     * @param defaultValue The default value to return if the given property is not defined.
     * @return The value of the given property, or the default value if the property is not defined
     */
    private <T> T getServerProperty(String property, Class<T> coerceTo, T defaultValue) {
        try {
            ConfigViewFactory factory = Services.getService(ConfigViewFactory.class);
            ConfigView view = factory.getView();
            ComposedConfigProperty<T> p = view.property(property, coerceTo);
            if (!p.isDefined()) {
                return defaultValue;
            }
            return p.get();
        } catch (OXException e) {
            logger.error("Error getting '{}' property. Returning the default value of '{}'", property, defaultValue, e);
        }
        return defaultValue;
    }

    
    private <T> T getContextSetProperty(String property, int userId, int contextId, Class<T> coerceTo, T defaultValue) {
        if (contextId == 0) {
            return defaultValue;
        }
        if (contextId < 0) {
            contextId = -contextId;
            userId = -1;
        }
        try {
            ConfigViewFactory factory = Services.getService(ConfigViewFactory.class);
            ConfigView view = factory.getView(userId, contextId);

            ConfigProperty<T> p = view.property("contextSets", property, coerceTo);
            if (!p.isDefined()) {
                return defaultValue;
            }

            return p.get();
        } catch (OXException e) {
            logger.error("Error getting '{}' property for user '{}' in context '{}'. Returning the default value of '{}'", property, userId, contextId, defaultValue, e);
        }
        return defaultValue;

    }
    /**
     * Get the value T of specified property for the specified user in the specified context and coerce it to the specified type T
     *
     * @param property The property's name
     * @param userId The user identifier
     * @param contextId The context identifier
     * @param coerceTo The type T to coerce the value of the property
     * @return The value T of the property
     * @throws OXException
     */
    private <T> T getUserProperty(String property, int userId, int contextId, Class<T> coerceTo, T defaultValue) {
        if (contextId == 0) {
            return defaultValue;
        }
        if (contextId < 0) {
            contextId = -contextId;
            userId = -1;
        }

        try {
            ConfigViewFactory factory = Services.getService(ConfigViewFactory.class);
            ConfigView view = factory.getView(userId, contextId);

            ComposedConfigProperty<T> p = view.property(property, coerceTo);
            if (!p.isDefined()) {
                return defaultValue;
            }

            return p.get();
        } catch (OXException e) {
            logger.error("Error getting '{}' property for user '{}' in context '{}'. Returning the default value of '{}'", property, userId, contextId, defaultValue, e);
        }
        return defaultValue;
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.config.Reloadable#getInterests()
     */
    @Override
    public Interests getInterests() {
        return DefaultInterests.builder().propertiesOfInterest("com.openexchange.guard.*").configFileNames("guard-core.properties", "server.properties").build();
    }
}
